//
// FILE:	map.C
// AUTHORS:	Emanuel G. Noik
//
//		Tcl/Tk implementation of dynamic mapping methods.

#include <assert.h>
#include <tk.h>

#include <util/MN_Dictionary.h>
#include <util/MN_Set.h>
#include <util/MN_Point.h>
#include <util/unix.h>
#include <hygraph/Hygraph.h>
#include <server/io.h>
#include <server/server.h>
#include <server/hgInfo.h>

#include "tclTkHInfo.h"
#include "tkHInfo.h"

//------------------------------------------------------------------------------

static int sort_index;

//------------------------------------------------------------------------------

static int DecreasingVal( const Vertex** u, const Vertex** v )
{
   float ua = DQ(val[sort_index],(*u));
   float va = DQ(val[sort_index],(*v));

   if( ua > va )	return( -1 );
   else if( ua < va )	return(  1 );
   else			return(  0 );
}

//------------------------------------------------------------------------------

PAttrLabel attributes[] = {
   { "rel x",			PA_RelPosX		},
   { "rel y",			PA_RelPosY		},
   { "rel z",			PA_RelPosZ		},
   { "abs x",			PA_AbsPosX		},
   { "abs y",			PA_AbsPosY		},
   { "rel size",		PA_RelSize		},
   { "rel x size",		PA_RelSizeX		},
   { "rel y size",		PA_RelSizeY		},
   { "abs x size",		PA_AbsSizeX		},
   { "abs y size",		PA_AbsSizeY		},
   { "hue",			PA_HSVh			},
   { "saturation",		PA_HSVs			},
   { "value",			PA_HSVv			},
   { "link width",		PA_LinkWidth		},
   { "fill percentage",		PA_FillPerc		},
   { "label",			PA_Label		},
   { 0,				PA_Label		}
};

//------------------------------------------------------------------------------

inline HGContainer* Cntr( HGObject* v )
{
   return v->c->FindSubContainer( REG_I );
}

//------------------------------------------------------------------------------

int TkHInfo::MapAttr( int t, int indx, PAttribute attr, char* attrLabel )
{
   if( sdirty[t][indx] ) ComputeStats( t, indx );

   double dfltWidth, dfltHeight;
   sprintf( tmpbuf, "expr $config(halfWidth,$config(iconType))" );
   Tcl_GlobalEval( interp, tmpbuf );
   dfltWidth = atof( interp->result ) * 2.0;
   sprintf( tmpbuf, "expr $config(halfHeight,$config(iconType))" );
   Tcl_GlobalEval( interp, tmpbuf );
   dfltHeight = atof( interp->result ) * 2.0;

   if( indx == -1 || stat[t][indx].Min() == stat[t][indx].Max() )
   {
      HGObject* o;
      for( int j = 0; (o = Object( t, j )) != NIL; j++ )
      {
         switch( attr )
         {
            case PA_RelPosX:
            case PA_RelPosY:
            case PA_RelPosZ:
            case PA_AbsPosX:
            case PA_AbsPosY:
               break;
            case PA_HSVh:
               if( t == 0 )
                  SetHSV( (Vertex*)o, 1, 0 );
               else
                  SetHSV( (AArc*)o,   1, 0 );
               break;
            case PA_HSVs:
               if( t == 0 )
                  SetHSV( (Vertex*)o, 1, 1 );
               else
                  SetHSV( (AArc*)o,   1, 1 );
               break;
            case PA_HSVv:
               if( t == 0 )
                  SetHSV( (Vertex*)o, 1, 2 );
               else
                  SetHSV( (AArc*)o,   1, 2 );
               break;
            case PA_RelSize:
            {
               HGContainer* c = Cntr(o);
               sprintf( tmpbuf, "ResizeNode %d $state(size)", c->widget );
               Tcl_GlobalEval( interp, tmpbuf );
               break;
            }
            case PA_RelSizeX:
            case PA_AbsSizeX:
            {
               HGContainer* c = Cntr( o );
               c->size.x = dfltWidth;
               SetSize( c );
               break;
            }
            case PA_RelSizeY:
            case PA_AbsSizeY:
            {
               HGContainer* c = Cntr( o );
               c->size.y = dfltHeight;
               SetSize( c );
               break;
            }
            case PA_FillPerc:
            {
               HGContainer* c = Cntr( o );
               sprintf( tmpbuf,
               ".canvas itemconfigure %d -perc 1", c->widget );
               Tcl_GlobalEval( interp, tmpbuf );
               break;
            }
            case PA_Label:
            {
               if( t == 0 )
               {
                  HGContainer* c = o->c->FindSubContainer( REG_L );
                  sprintf( tmpbuf, "SetLabel %d {%s}",
                     c->widget, hygraph->Label( (Vertex*)o ) );
                  Tcl_GlobalEval( interp, tmpbuf );
                  int dummy;
                  sscanf( interp->result, "%d %g %g", &dummy,
                     &c->size.x, &c->size.y );
               }
               break;
            }
            case PA_LinkWidth:
            {
               sprintf( tmpbuf, ".canvas itemconfigure %d -width 1",
                  ((AArc*)o)->ca->widget );
               Tcl_GlobalEval( interp, tmpbuf );
               break;
            }
            default:
               break;
         }
      }
   }
   else   // indx != -1
   {
      Tcl_GlobalEval( interp, "winfo width .canvas" );
      int width = atoi( interp->result );

      Tcl_GlobalEval( interp, "winfo height .canvas" );
      int height = atoi( interp->result );

      float minSize = atof( Tcl_GetVar2( interp, "state", "dm,relsize,min",
                            TCL_GLOBAL_ONLY ) );
      float maxSize = atof( Tcl_GetVar2( interp, "state", "dm,relsize,max",
                            TCL_GLOBAL_ONLY ) );
      int maxWidth = atoi( Tcl_GetVar2( interp, "state", "dm,linkwidth,max",
                            TCL_GLOBAL_ONLY ) );

      width  -= 20;
      height -= 20;

      if( attr == PA_RelPosZ )
      {
         sort_index = indx;
         hygraph->SortV( (ComparatorFunction)DecreasingVal );
      }

      HGObject* o;
      for( int j = 0; (o = Object( t, j )) != NIL; j++ )
      {
         switch( attr )
         {
            case PA_RelPosX:
            {
               HGContainer* ci = o->c->FindSubContainer( REG_I );

               int x =
                  MapNormal(10,width,stat[t][indx].Normal(DQ(val[indx],o)));
               //sprintf( tmpbuf, ".canvas coords %d", ci->widget );
               //Tcl_GlobalEval( interp, tmpbuf );
               {
                  //float xx, yy, u, v;
                  float xx = ci->pos.x + ci->offset.x;
                  MN_Point d;
                  //sscanf( interp->result, "%g %g %g %g", &xx, &yy, &u, &v );
                  if( DQ(masked,o) )
                     d.Set( x - Round(xx) + 10000, 0, 0 );
                  else
                     d.Set( x - Round(xx), 0, 0 );
                  Move( (Vertex*)o, &d );
               }
               Tcl_GlobalEval( interp, tmpbuf );
               break;
            }
            case PA_RelPosY:
            {
               HGContainer* ci = o->c->FindSubContainer( REG_I );

               int y =
                  MapNormal(10,height,1-stat[t][indx].Normal(DQ(val[indx],o)));
               //sprintf( tmpbuf, ".canvas coords %d", ci->widget );
               //Tcl_GlobalEval( interp, tmpbuf );
               {
                  //float xx, yy, u, v;
                  float yy = ci->pos.y + ci->offset.y;
                  MN_Point d;
                  //sscanf( interp->result, "%g %g %g %g", &xx, &yy, &u, &v );
                  if( DQ(masked,o) )
                     d.Set( 0, y - Round(yy) + 10000, 0 );
                  else
                     d.Set( 0, y - Round(yy), 0 );
                  Move( (Vertex*)o, &d );
               }
               Tcl_GlobalEval( interp, tmpbuf );
               break;
            }
            case PA_RelPosZ:
            {
               HGContainer* cl = o->c->FindSubContainer( REG_L );
               HGContainer* ci = o->c->FindSubContainer( REG_I );
               sprintf( tmpbuf, ".canvas lower %d", cl->widget );
               Tcl_GlobalEval( interp, tmpbuf );
               sprintf( tmpbuf, ".canvas lower %d", ci->widget );
               Tcl_GlobalEval( interp, tmpbuf );
               sprintf( tmpbuf, ".canvas lower %d", o->c->widget );
               Tcl_GlobalEval( interp, tmpbuf );
               break;
            }
            case PA_AbsPosX:
            {
               HGContainer* c = Cntr( o );
               MN_Point d( DQ(val[indx],o) - c->pos.x, 0, 0 );
               Move( (Vertex*)o, &d );
               break;
            }
            case PA_AbsPosY:
            {
               HGContainer* c = Cntr( o );
               MN_Point d( 0, DQ(val[indx],o) - c->pos.y, 0 );
               Move( (Vertex*)o, &d );
               break;
            }
            case PA_RelSize:
            {
               HGContainer* c = Cntr( o );
               static char* size[] = { "Small", "Normal", "Large", 0 };
               float s = minSize + (maxSize-minSize) *
                  stat[t][indx].Normal( DQ(val[indx],o) );
               int i = s < 1 ? 0 : (s < 1.5 ? 1 : 2);
               sprintf( tmpbuf, "ResizeNode %d %s %g",
                  c->widget, size[i], sqrt(s) );
               Tcl_GlobalEval( interp, tmpbuf );
               break;
            }
            case PA_RelSizeX:
            {
               HGContainer* c = Cntr( o );
               float s = minSize + (maxSize-minSize) *
                  stat[t][indx].Normal(DQ(val[indx],o));
               c->size.x = dfltWidth * s;
               SetSize( c );
               break;
            }
            case PA_RelSizeY:
            {
               HGContainer* c = Cntr( o );
               float s = minSize + (maxSize-minSize) *
                  stat[t][indx].Normal(DQ(val[indx],o));
               c->size.y = dfltHeight * s;
               SetSize( c );
               break;
            }
            case PA_AbsSizeX:
            {
               HGContainer* c = Cntr( o );
               c->size.x = Max(DQ(val[indx],o),(float)0.0);
               SetSize( c );
               break;
            }
            case PA_AbsSizeY:
            {
               HGContainer* c = Cntr( o );
               c->size.y = Max(DQ(val[indx],o),(float)0.0);
               SetSize( c );
               break;
            }
            case PA_HSVh:
               if( t == 0 )
                  SetHSV( (Vertex*)o,stat[t][indx].Normal(DQ(val[indx],o)),0);
               else
                  SetHSV( (AArc*)o,  stat[t][indx].Normal(DQ(val[indx],o)),0);
               break;
            case PA_HSVs:
               if( t == 0 )
                  SetHSV( (Vertex*)o, stat[t][indx].Normal(DQ(val[indx],o)),1);
               else
                  SetHSV( (AArc*)o,   stat[t][indx].Normal(DQ(val[indx],o)),1);
               break;
            case PA_HSVv:
               if( t == 0 )
                  SetHSV( (Vertex*)o, stat[t][indx].Normal(DQ(val[indx],o)),2);
               else
                  SetHSV( (AArc*)o,   stat[t][indx].Normal(DQ(val[indx],o)),2);
               break;
            case PA_FillPerc:
            {
               HGContainer* c = Cntr( o );
               sprintf( tmpbuf,
               ".canvas itemconfigure %d -perc %g", c->widget,
                  stat[t][indx].Normal(DQ(val[indx],o)) );
               Tcl_GlobalEval( interp, tmpbuf );
               break;
            }
            case PA_Label:
            {
               if( t == 0 )
               {
                  HGContainer* c = o->c->FindSubContainer( REG_L );
                  sprintf( tmpbuf, "SetLabel %d {%g}",
                     c->widget, DQ(val[indx],o) );
                  Tcl_GlobalEval( interp, tmpbuf );
                  int dummy;
                  sscanf( interp->result, "%d %g %g", &dummy,
                     &c->size.x, &c->size.y );
               }
               break;
            }
            case PA_LinkWidth:
            {
               int i = MapNormal( 1, maxWidth,
                  stat[t][indx].Normal(DQ(val[indx],o)));
               sprintf( tmpbuf, ".canvas itemconfigure %d -width %d",
                        ((AArc*)o)->ca->widget, i );
               Tcl_GlobalEval( interp, tmpbuf );
               break;
            }
            default:
               break;
         }
      }
   }

   switch( attr )
   {
      case PA_RelSize:
      case PA_RelSizeX:
      case PA_RelSizeY:
      case PA_Label:
         Tcl_GlobalEval( interp, "l nop 1" );
         break;
      default:
         break;
   }

   // EGN: why am I doing this?  A: To remember mapping for remaping...
   sprintf( tmpbuf, "set config([format \"%cattr,%s\"]) %d",
      Code( t ), attrLabel, indx );
   Tcl_GlobalEval( interp, tmpbuf );

   return ERR_OK;
}

//------------------------------------------------------------------------------

static void HSV( float s, MN_Point* p, int component )
{
   switch( component )
   {
      case 0:		// hue
         p->x = (1.0 - s) * 297.0;
         if( p->y == 0 ) p->y = 1.0;
         if( p->z == 0 ) p->z = 0.9;
         break;
      case 1:		// saturation
         p->y = 0.19 + 0.81 * s;
         if( p->z == 0 ) p->z = 0.9;
         break;
      case 2:		// value
         p->z = 0.4 + 0.5 * s;
         if( p->y == 0 ) p->y = 1.0;
         break;
      default:
         if( p->y == 0 ) p->y = 1.0;
         if( p->z == 0 ) p->z = 0.9;
         break;
   }
}

//------------------------------------------------------------------------------

void TkHInfo::SetHSV( Vertex* v, float s, int component )
{
   MN_Point rgb;
   char rgbf[16], rgbt[16];
   // EGN: don't use c->gc any more. Instead, add an MN_Color
   //  object to DQInfo
   //
   HGContainer* c = Cntr( v );
   HSV( s, &( c->gc.bg_color ), component );
   HSVtoRGB( &( c->gc.bg_color ), &rgb );

   // Compute fill color string.
   //
   sprintf( rgbf, "#%02x%02x%02x",
            (int)(rgb.x * 255), (int)(rgb.y * 255), (int)(rgb.z * 255) );

   // Compute text color string.
   // Hue: [0-45,200-]
   //         Saturation: > 0.5 white else black
   //      [46-199] black
   //
   if( c->gc.bg_color.x > 45 && c->gc.bg_color.x < 200 )
      sprintf( rgbt, "#000000" );
   else
   {
      if( c->gc.bg_color.y > 0.5 )
         sprintf( rgbt, "#FFFFFF" );
      else
         sprintf( rgbt, "#000000" );
   }

   // Set colors.
   //
/*
   sprintf( tmpbuf,
            "$fFromId(%s).l configure -bg %s -fg %s",
            //"$fFromId(%s).l configure -bg %s",
            hygraph->Id(v), rgbf, rgbt );
            //hygraph->Id(v), rgbf );
*/
   if( DQ(painted,v) )
      sprintf( DQ(color,v), rgbf );
   else
   {
      sprintf( tmpbuf, "NodeColor %d %s %s", c->widget, rgbf, rgbt );
      Tcl_GlobalEval( interp, tmpbuf );
   }
}

void TkHInfo::SetHSV( AArc* a, float s, int component )
{
   MN_Point rgb;
   char rgbf[16];

   HSV( s, &( a->ca->gc.bg_color ), component );

   HSVtoRGB( &( a->ca->gc.bg_color ), &rgb );
   sprintf( rgbf, "#%02x%02x%02x",
            (int)(rgb.x * 255), (int)(rgb.y * 255), (int)(rgb.z * 255) );

   if( DQ(painted,a) )
      sprintf( DQ(color,a), rgbf );
   else
   {
      sprintf( tmpbuf, ".canvas itemconfigure %d -fill %s", a->ca->widget,
         rgbf );
      Tcl_GlobalEval( interp, tmpbuf );
   }
}

//------------------------------------------------------------------------------

void TkHInfo::SetSize( HGContainer* c )
{
   sprintf( tmpbuf, ".canvas coords %d %g %g %g %g",
      c->widget,
      c->pos.x + c->offset.x,
      c->pos.y + c->offset.y,
      c->pos.x + c->offset.x + Max( c->size.x, (float)0.0 ),
      c->pos.y + c->offset.y + Max( c->size.y, (float)0.0 ) );
   Tcl_GlobalEval( interp, tmpbuf );
}

//------------------------------------------------------------------------------
